package zero1hd.rhythmbullet.audio;

import java.util.Observable;
import java.util.Observer;

import com.badlogic.gdx.files.FileHandle;
import com.badlogic.gdx.utils.Array;
import com.badlogic.gdx.utils.Disposable;

import zero1hd.rhythmbullet.audio.metadata.AudioMetadata;
import zero1hd.rhythmbullet.audio.metadata.MP3Metadata;
import zero1hd.rhythmbullet.audio.metadata.WAVMetadata;

public class MusicMetadataController extends Observable implements Disposable, Observer {
	private MusicList musicList;
	private Array<AudioMetadata> metadataArray;
	private MetadataLoadingThread loadingThread;
	private volatile boolean searching;
	
	public MusicMetadataController(MusicList musicList) {
		this.musicList = musicList;
		metadataArray = new Array<>();
		loadingThread = new MetadataLoadingThread();
		musicList.addObserver(this);
	}
	
	public MusicList getMusicList() {
		return musicList;
	}
	
	/**
	 * Non-blocking, loads on separate thread.
	 */
	public void loadAudioMetadata() {
		if (!loadingThread.start()) {
			loadingThread.stop();
			loadingThread = new MetadataLoadingThread();
			loadingThread.start();
		}
	}
	
	public boolean isDone() {
		return (metadataArray.size == musicList.getMusicArray().size);
	}
	
	@Override
	public void dispose() {
		for (int i = 0; i < metadataArray.size; i++) {
			metadataArray.get(i).dispose();
		}
	}
	
	public int size() {
		synchronized (loadingThread) {
			return metadataArray.size;
		}
	}
	
	public AudioMetadata getMetadata(int index) {
		synchronized (loadingThread) {
			return metadataArray.get(index);
		}
	}
	
	
	public AudioMetadata getInfo(FileHandle filehandle) {
		return metadataArray.get(musicList.getMusicArray().indexOf(filehandle, true));
	}
	
	public boolean isSearching() {
		return searching;
	}
	
	private class MetadataLoadingThread implements Runnable {
		private Thread thread;
		private String name = "Metadata-Load";
		private volatile boolean work;
		
		@Override
		public void run() {
			searching = true;
			for (int i = 0; i < metadataArray.size; i++) {
				metadataArray.get(i).dispose();
			}
			metadataArray.clear();
			for (int i = 0; i < musicList.getTotal() && work; i++) {
				FileHandle musicFile = musicList.getMusicArray().get(i);
				synchronized (this) {
					switch (SupportedFormats.valueOf(musicFile.extension().toUpperCase())) {
					case MP3:
						metadataArray.add(new MP3Metadata(musicFile));
						break;
					case WAV:
						metadataArray.add(new WAVMetadata(musicFile));
						break;
					default:
						break;
					
					}
				}
			}
			if (work) {
				searching = false;
				notifyObservers();
			}
		}
		
 		public boolean start() {
 			if (thread == null) {
 				thread = new Thread(this, name);
 				thread.start();
 				return true;
 			} else {
 				return false;
 			}
 		}
 		
 		public void stop() {
 			work = false;
 		}
	}

	@Override
	public void update(Observable o, Object arg) {
		if (o == musicList) {
			loadingThread.stop();
			loadAudioMetadata();
		}
	}
}
