package zero1hd.rhythmbullet.audio.analyzer;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.math.MathUtils;
import com.badlogic.gdx.utils.FloatArray;

import edu.emory.mathcs.jtransforms.fft.FloatFFT_1D;
import zero1hd.rhythmbullet.audio.processor.AudioProcessor;

public class AudioAnalyzer implements Runnable {
	private Thread thread;
	private String threadName = "Audio-Analyzer";
	
	private volatile boolean work = true;
	private int windowSize = 1024;
	
	private FloatArray bassSpectralFlux = new FloatArray();
	private FloatArray mSpectralFlux = new FloatArray();
	private FloatArray umSpectralFlux = new FloatArray();

	private FloatArray bassThreshold = new FloatArray();
	private FloatArray mThreshold = new FloatArray();
	private FloatArray umThreshold = new FloatArray();
	
	private FloatArray bassPrunned = new FloatArray();
	private FloatArray mPrunned = new FloatArray();
	private FloatArray umPrunned = new FloatArray();
	
	private FloatArray bassPeaks = new FloatArray();
	private FloatArray mPeaks = new FloatArray();
	private FloatArray umPeaks = new FloatArray();
	
	private float bassMaxValue, mMaxValue, umMaxValue, secondsPerWindow, mAvg, bassAvg, umAvg;
	
	AudioProcessor processor;
	
	private int PUID;
	private int progress;
	
	public boolean start(AudioProcessor processor) {
		if (thread == null) {
			this.processor = processor;
			work = true;
			thread = new Thread(this, threadName);
			thread.start();
			return true;
		} else {
			return false;
		}
	}
	
	@Override
	public void run() {
		spectralFluxAnalysis();
		thresholdCalculation();
		pruneFluxValues();
		peakDetection();
	}
	
	private void spectralFluxAnalysis() {
		progress = 0;
		int tasksDone = 0;
		long totalTasks = MathUtils.round((float)processor.getSampleFrames()/windowSize);
		
		float[] audioPCM = new float[windowSize];
		float[] spectrum = new float[(windowSize/2)+1];
		float[] lastSpectrum = new float[(windowSize/2)+1];
		
		int bassBinBegin = 1;
		int bassBinEnd = 11;
		
		int mBinBegin = 50;
		int mBinEnd = 250;
		
		int umBinBegin = 350;
		int umBinEnd = 513;
		
		Gdx.app.debug("Read freq", String.valueOf(processor.getSampleRate()));
		Gdx.app.debug("Using following bin ranges", "\nBass freq begin: " + bassBinBegin + "\nBass freq end: " + bassBinEnd + "\nMain freq begin: " + umBinBegin + "\nMain freq end: " + umBinEnd);

		Gdx.app.debug("Total tasks", String.valueOf(totalTasks));
		
		FloatFFT_1D fft = new FloatFFT_1D(windowSize);
		int seedDigit = 0;
		
		while (processor.readFrames(audioPCM) > 0 && work) {
			
			fft.realForward(audioPCM);
			
			//Building a PUID (Pseudo unique ID)
			if (tasksDone == (seedDigit*totalTasks/9)) {
				float avg = 0;
				for (int frame = 0; frame < spectrum.length; frame++) {
					avg += spectrum[frame];
				}
				avg /= spectrum.length;
				if (avg < 0) {
					avg *= -1f;
				}
				PUID +=(int) Math.pow(10, 9-seedDigit) * ((int)(avg*1000f)-(int)(avg*100f)*10);
				seedDigit ++;
			}
			
			System.arraycopy(spectrum, 0, lastSpectrum, 0, spectrum.length);
			System.arraycopy(audioPCM, 0, spectrum, 0, spectrum.length);
			
			float fluxVal;
			//bass detection
			fluxVal = 0;
			for (int i = bassBinBegin; i < bassBinEnd; i++) {
				fluxVal += ((spectrum[i] - lastSpectrum[i])) < 0
						? 0 : (spectrum[i] - lastSpectrum[i]);
			}
			bassSpectralFlux.add(fluxVal);
			
			//m detection
			fluxVal = 0;
			for (int i = mBinBegin; i < mBinEnd; i++) {
				fluxVal += ((spectrum[i] - lastSpectrum[i])) < 0
						? 0 : (spectrum[i] - lastSpectrum[i]);
			}
			mSpectralFlux.add(fluxVal);
			
			//um detection
			fluxVal = 0;
			for (int i = umBinBegin; i < umBinEnd; i++) {
				fluxVal += ((spectrum[i] - lastSpectrum[i])) < 0
						? 0 : (spectrum[i] - lastSpectrum[i]);
			}
			umSpectralFlux.add(fluxVal);
			tasksDone++;
			progress = (int) (100f*tasksDone/totalTasks);
		}
		
		if (work) {
			Gdx.app.debug("Audio Analyzer", "Done getting spectral flux.");
			Gdx.app.debug("Audio Analyzer", "window count: " + bassSpectralFlux.size);
			Gdx.app.debug("Audio Analyzer", "USING SEED: " + PUID);
			progress = 100;
		}
	}
	
	private void thresholdCalculation() {
		Gdx.app.debug("Audio Analyzer", "beginning threshold calc.");

		float bassThresholdMultiplier = 1.5f;
		float mThresholdMultiplier = 1.4f;
		float umThresholdMultiplier = 1.4f;
		
		int bassThresholdCalcRange = (int) (0.27/(windowSize/processor.getSampleRate()));
		int mThresholdCalcRange = (int) (0.4/(windowSize/processor.getSampleRate()));
		int umThresholdCalcRange = (int) (0.4/(windowSize/processor.getSampleRate()));
		//threshold calculation
		for (int i = 0; i < umSpectralFlux.size && work; i++) {
			int start = Math.max(0, i - bassThresholdCalcRange/2);
			int end = Math.min(umSpectralFlux.size - 1, i + bassThresholdCalcRange/2);
			float average = 0;
			for (int j = start; j <= end; j++) {
				average += bassSpectralFlux.get(j);
			}
			average /= (end - start);
			bassThreshold.add(average * bassThresholdMultiplier);
			
			start = Math.max(0, i - mThresholdCalcRange/2);
			end = Math.min(umSpectralFlux.size - 1, i + mThresholdCalcRange/2);
			average = 0;
			for (int j = start; j <= end; j++) {
				average+= mSpectralFlux.get(j);
			}
			average /= (end - start);
			mThreshold.add(average*mThresholdMultiplier);
			
			start = Math.max(0, i - umThresholdCalcRange/2);
			end = Math.min(umSpectralFlux.size - 1, i + umThresholdCalcRange/2);
			average = 0;
			for (int j = start; j <= end; j++) {
				average+= umSpectralFlux.get(j);
			}
			average /= (end - start);
			umThreshold.add(average*umThresholdMultiplier);
		}
		Gdx.app.debug("Audio Analyzer", "Threshold calculated.");
	}
	
	private void pruneFluxValues() {
		//pruning data
		float prunnedCurrentVal;

		for (int i = 0; i < umSpectralFlux.size && work; i++) {
			prunnedCurrentVal = bassSpectralFlux.get(i) - bassThreshold.get(i);
			if (prunnedCurrentVal >= 0) {
				bassPrunned.add(prunnedCurrentVal);
			} else {
				bassPrunned.add(0);
			}
			
			prunnedCurrentVal = mSpectralFlux.get(i) - mThreshold.get(i);
			if (prunnedCurrentVal >= 0 ) {
				mPrunned.add(prunnedCurrentVal);
			} else {
				mPrunned.add(0);
			}
			
			prunnedCurrentVal = umSpectralFlux.get(i) - umThreshold.get(i);
			if (prunnedCurrentVal >= 0 ) {
				umPrunned.add(prunnedCurrentVal);
			} else {
				umPrunned.add(0);
			}
		}
		Gdx.app.debug("Audio Analyzer", "Data prunned.");
	}
	
	private void peakDetection() {
		int lastBeatID = 0;
		float bassBeats = 0;
		float mBeats = 0;
		float umBeats = 0;
		float avgSPB = -1f;

		for (int i = 0; i < umPrunned.size-1 && work; i++) {
			bassPeaks.add((bassPrunned.get(i) > bassPrunned.get(i+1) ? bassPrunned.get(i) : 0f));
			if (bassPeaks.get(i) > bassMaxValue) {
				bassMaxValue = bassPeaks.get(i);
			}
			
			mPeaks.add((mPrunned.get(i) > mPrunned.get(i+1) ? mPrunned.get(i) : 0f));
			if (mPeaks.get(i) > mMaxValue) {
				mMaxValue = mPeaks.get(i);
			}
			
			umPeaks.add((umPrunned.get(i) > umPrunned.get(i+1) ? umPrunned.get(i) : 0f));
			if (umPeaks.get(i) > umMaxValue) {
				umMaxValue = umPeaks.get(i);
			}
			

			if (avgSPB != -1) {
				if (bassPeaks.get(i) == 0) {
					avgSPB ++;
				} else {
					lastBeatID = i;
				}
			} else if (bassPeaks.get(i) != 0) {
				avgSPB = 0;
			}

			if (bassPeaks.get(i) != 0) {
				bassAvg += bassPeaks.get(i);
				bassBeats++;
			}
			
			if (mPeaks.get(i) != 0) {
				mAvg += mPeaks.get(i);
				mBeats++;
			}
			
			if (umPeaks.get(i) != 0) {
				umAvg += umPeaks.get(i);
				umBeats++;
			}
		}
		
		secondsPerWindow = windowSize/processor.getSampleRate();

		//then we minus one from the beats so it actually works out
		avgSPB -= bassPrunned.size-lastBeatID;
		avgSPB *= secondsPerWindow;
		avgSPB /= bassBeats;
		Gdx.app.debug("Audio Analyzer", "Avg SPB: " + avgSPB);

		bassAvg /= bassBeats;
		mAvg /= mBeats;
		umAvg /= umBeats;
		Gdx.app.debug("Audio Analyzer", "Avg bass: " + bassAvg);
		Gdx.app.debug("Audio Analyzer", "Avg M: " + mAvg);
		Gdx.app.debug("Audio Analyzer", "Avg UM: " + umAvg);
		
	}
	
	public int getProgress() {
		return progress;
	}
	
	public void stop() {
		work = false;
	}
}
