package zero1hd.rhythmbullet.audio.visualizer;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.graphics.Color;
import com.badlogic.gdx.graphics.GL20;
import com.badlogic.gdx.graphics.g2d.Batch;
import com.badlogic.gdx.graphics.glutils.ShapeRenderer;
import com.badlogic.gdx.graphics.glutils.ShapeRenderer.ShapeType;
import com.badlogic.gdx.math.MathUtils;
import com.badlogic.gdx.utils.Disposable;

import zero1hd.rhythmbullet.audio.MusicController;

public class DoubleHorizontalVisualizer implements Disposable {
	private int width, height, barWidth, spaceBetweenBars;
	private int x, y;
	private ShapeRenderer shapeRenderer;
	private PCMSystem pcm;
	private float[] amplitudes;
	private int[] barHeights;
	private int binsPerBar;
	private float offset;
	private int boundaryThickness;
	private float baseSensitivity;
	private float targetDelta;
	private float spacePercentage = 0.7f;
	private int barCount = 90;
	private float barChangeRate = 14f;
	private int smoothRange = 1;
	private int binsToInclude = 180;
	private Color color = new Color(0.5f, 0.6f, 0.8f, 0.46f);
	/**
	 * 
	 * @param barCount amount of bars this visualizer should have.
	 * @param width the width of the visualizer.
	 * @param spacePercentage the percentage of a bar that should be space.
	 */
	public DoubleHorizontalVisualizer(int width, int height, float baseSensitivity, int boundaryThickness, int targetFPS, MusicController musicController, PCMSystem PCMSystem) {
		this.barWidth = width/barCount;
		this.spaceBetweenBars = MathUtils.round(barWidth * spacePercentage);
		this.barWidth -= spaceBetweenBars;
		this.baseSensitivity = baseSensitivity;
		pcm = PCMSystem;
		binsPerBar = (binsToInclude/barCount);
		this.width = width;
		this.height = height;
		amplitudes = new float[barCount];
		barHeights = new int[barCount];
		shapeRenderer = new ShapeRenderer();
		boundaryThickness = barWidth;
		offset = (width - (barCount*(barWidth+spaceBetweenBars)-spaceBetweenBars))/2f + x;
		this.targetDelta = 1f/targetFPS;
	}
	
	public void act(float delta) {
		float[] freqBins = pcm.getFrequencyBins();
		for (int bar = 0; bar < amplitudes.length; bar++) {
			float normalizedAmplitude = 0;
			for (int freq = bar*binsPerBar; freq < (bar*binsPerBar) + binsPerBar; freq++) {
				normalizedAmplitude += Math.abs(freqBins[freq]);
			}
			amplitudes[bar] += normalizedAmplitude * baseSensitivity;
			amplitudes[bar] /= binsPerBar;
		}
		for (int bar = 0; bar < barHeights.length; bar++) {
			int smoothCount = 1;
			for (int range = 0; range < smoothRange; range++) {
				if (bar+range < amplitudes.length) {
					amplitudes[bar] += amplitudes[bar+range];
					smoothCount++;
				}
				if (bar-range > 0) {
					amplitudes[bar] += amplitudes[bar-range];
					smoothCount++;
				}
			}
			amplitudes[bar] /= smoothCount;
			
			
			int pixelsMoved = 0;
			pixelsMoved = MathUtils.round(amplitudes[bar] - barHeights[bar]);
			pixelsMoved = MathUtils.floor(pixelsMoved*targetDelta*barChangeRate);
			
			
			barHeights[bar] += pixelsMoved;
			
			if (barHeights[bar] < 0) barHeights[bar] = 0;
		}
	}
	
	public void draw(Batch batch, float parentAlpha) {
		batch.end();
		
		Gdx.gl.glEnable(GL20.GL_BLEND);
		shapeRenderer.begin(ShapeType.Filled);
		shapeRenderer.setProjectionMatrix(batch.getProjectionMatrix());
		shapeRenderer.setTransformMatrix(batch.getTransformMatrix());
		if (boundaryThickness > 0) {
			shapeRenderer.rect(x, y, width, boundaryThickness);
			shapeRenderer.rect(x, y+height, width, -boundaryThickness);
		}
		
		for (int bar = 0; bar < barCount; bar++) {
			shapeRenderer.setColor(color);
			shapeRenderer.rect(offset + (spaceBetweenBars+barWidth)*bar, y+height, barWidth, barHeights[bar]);
			shapeRenderer.rect(offset + (spaceBetweenBars+barWidth)*bar, y-barHeights[barHeights.length - 1 - bar], barWidth, barHeights[barHeights.length - 1 - bar]);
		}
		shapeRenderer.end();
		Gdx.gl.glDisable(GL20.GL_BLEND);
		batch.begin();
	}
	
	public int getWidth() {
		return width;
	}
	
	public int getHeight() {
		return height;
	}
	
	public void setX(int x) {
		this.x = x;
	}
	
	public void setY(int y) {
		this.y = y;
	}
	
	public void setPosition(int x, int y) {
		this.x = x;
		this.y = y;
	}
	
	public int getX() {
		return x;
	}
	
	public int getY() {
		return y;
	}
	
	@Override
	public void dispose() {
		pcm.dispose();
	}
}
