package zero1hd.rhythmbullet.desktop.screens.main;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.Input.Keys;
import com.badlogic.gdx.ScreenAdapter;
import com.badlogic.gdx.graphics.GL20;
import com.badlogic.gdx.graphics.Texture;
import com.badlogic.gdx.graphics.g2d.Batch;
import com.badlogic.gdx.graphics.g2d.SpriteBatch;
import com.badlogic.gdx.math.Vector3;
import com.badlogic.gdx.scenes.scene2d.Actor;
import com.badlogic.gdx.scenes.scene2d.InputEvent;
import com.badlogic.gdx.scenes.scene2d.InputListener;
import com.badlogic.gdx.scenes.scene2d.Stage;
import com.badlogic.gdx.scenes.scene2d.utils.ChangeListener;
import com.badlogic.gdx.scenes.scene2d.utils.ClickListener;
import com.badlogic.gdx.utils.viewport.ScreenViewport;

import zero1hd.rhythmbullet.RhythmBullet;
import zero1hd.rhythmbullet.audio.MusicMetadataController;
import zero1hd.rhythmbullet.audio.MusicList;
import zero1hd.rhythmbullet.audio.MusicController;
import zero1hd.rhythmbullet.desktop.audio.processor.DesktopAudioProcessorFactory;
import zero1hd.rhythmbullet.graphics.shaders.BloomShader;
import zero1hd.rhythmbullet.graphics.ui.Page;
import zero1hd.rhythmbullet.util.ResizeReadyScreen;

public class MainScreen extends ScreenAdapter implements ResizeReadyScreen {
	private boolean resizing;
	
	private Stage stage;
	private Vector3 cameraPosition;
	private Listeners listeners;
	
	private MainPage mainPage;
	private OptionsPage optionsPage;
	private KeybindPage keybindPage;
	private GraphicsPage graphicsPage;
	private CreditsPage creditsPage;
	private MusicSelectionPage musicSelectionPage;
	private AnalysisPage analysisPage;
	
	private MusicController musicController;
	private MusicMetadataController musicMetadataController;
	
	private RhythmBullet rhythmBullet;
	
	private BloomShader bloomShader;
	
	private float lerpAlpha;
	
	private Texture background;
	
	private Batch screenBatch;
	
	public MainScreen(RhythmBullet core) {
		this.rhythmBullet = core;
		stage = new Stage(new ScreenViewport());
		cameraPosition = new Vector3(stage.getCamera().position);
		
		MusicList musicList = new MusicList(new DesktopAudioProcessorFactory(), core.getPreferences().getString("music dir"));
		musicController = new MusicController(musicList, core.getPreferences());
		musicController.setAutoPlay(true);
		musicController.setShuffle(true);
		
		musicMetadataController = new MusicMetadataController(musicList);
		
		listeners = new Listeners();
		screenBatch = new SpriteBatch();
	}
	
	@Override
	public void render(float delta) {
		Gdx.gl.glClearColor(0f, 0f, 0f, 1f);
        Gdx.gl.glClear(GL20.GL_COLOR_BUFFER_BIT);
		stage.act(delta);
		if (bloomShader != null) {
			bloomShader.begin();
			draw();
			bloomShader.end(stage.getWidth(), stage.getHeight());
		} else {
			draw();
		}

		if (stage.getCamera().position.x != cameraPosition.x || stage.getCamera().position.y != cameraPosition.y) {
			stage.getCamera().position.lerp(cameraPosition, delta*lerpAlpha);
			stage.getViewport().apply();
		}
		
		super.render(delta);
	}
	
	private void draw() {
		if (!resizing) {
			stage.getViewport().apply();
			screenBatch.begin();
			screenBatch.draw(background, 0, 0, stage.getViewport().getScreenWidth(), stage.getViewport().getScreenHeight());
			screenBatch.end();
			stage.draw();
		}
	}
	
	@Override
	public void preAssetLoad() {
		resizing = true;
		stage.clear();
		if (bloomShader != null) {
			bloomShader.dispose();
			bloomShader = null;
		}
		background = null;
		musicController.deleteObservers();
	}
	
	@Override
	public void postAssetLoad() {
		
		background = rhythmBullet.getAssetManager().get("backgrounds/mainBG.png", Texture.class);
		
		mainPage = new MainPage(musicController, rhythmBullet.getAssetManager(), rhythmBullet.getSkin(), rhythmBullet.getScreenConfiguration(), listeners.musicSelectionPageButtonListener, listeners.optionsPageButtonListener);
		stage.addActor(mainPage);
		//End main menu

		optionsPage = new OptionsPage(musicController, rhythmBullet.getSkin(), rhythmBullet.getPreferences(), listeners.returnToMainPageListener, listeners.graphicsPageButtonListener, listeners.keybindPageButtonListener);
		stage.addActor(optionsPage);

		keybindPage = new KeybindPage(rhythmBullet.getAssetManager(), rhythmBullet.getSkin(), listeners.optionsPageButtonListener);
		stage.addActor(keybindPage);

		graphicsPage = new GraphicsPage(rhythmBullet.getSkin(), rhythmBullet.getPreferences(), listeners.optionsPageButtonListener, listeners.bloomLevelShaderListener);
		stage.addActor(graphicsPage);
		
		creditsPage = new CreditsPage(rhythmBullet.getSkin());
		stage.addActor(creditsPage);
		
		musicSelectionPage = new MusicSelectionPage(rhythmBullet.getAssetManager(), rhythmBullet.getSkin(), musicController, musicMetadataController, listeners.returnToMainPageListener, listeners.analysisPageButtonListener);
		stage.addActor(musicSelectionPage);
		
		analysisPage = new AnalysisPage(rhythmBullet.getSkin(), listeners.returnToMainPageListener, listeners.confirmedSongListener);
		stage.addActor(analysisPage);
		
		
		
		stage.addListener(new InputListener() {
			@Override
			public boolean keyUp(InputEvent event, int keycode) {
				if (keycode == Keys.ESCAPE) {
					stage.unfocusAll();
				}
				return super.keyUp(event, keycode);
			}
		});	
		
		stage.addListener(new ClickListener() {
			@Override
			public void clicked(InputEvent event, float x, float y) {
				if (stage.hit(x, y, true) == null) {
					stage.unfocusAll();
				}
				super.clicked(event, x, y);
			}
		});
		
		musicController.addObserver(musicSelectionPage);
		musicController.addObserver(mainPage);
		musicController.getMusicList().asyncSearch(false);
		resizing = false;
	}
	
	@Override
	public void show() {
		Gdx.input.setInputProcessor(stage);
		calcLerpAlpha(Gdx.graphics.getWidth());
		super.show();
	}
	
	@Override
	public void hide() {
		saveAll();
		super.hide();
	}
	
	public void saveAll() {
		if (optionsPage != null) {
			optionsPage.saveOptions();
			graphicsPage.saveOptions();
			rhythmBullet.getPreferences().flush();
		}
	}
	
	@Override
	public void resize(int width, int height) {
		stage.getViewport().update(width, height, false);
		cameraPosition.x = width/2;
		cameraPosition.y = height/2;
		calcLerpAlpha(width);
		super.resize(width, height);
	}
	
	@Override
	public void dispose() {
		Gdx.app.debug("Mainscreen", "disposing...");
		stage.dispose();
		musicMetadataController.dispose();
		screenBatch.dispose();
		super.dispose();
	}
	
	private void calcLerpAlpha(int width) {
		if (width <= 3835) {
			lerpAlpha = 5.0f;
		} else {
			lerpAlpha = 5.5f;
		}
	}
	
	public void setDisplayedPage(Page page) {
		page.setCameraPositionToPage(cameraPosition);
	}
	
	private class Listeners {
		ChangeListener musicSelectionPageButtonListener = new ChangeListener() {
			@Override
			public void changed(ChangeEvent event, Actor actor) {
				setDisplayedPage(musicSelectionPage);
				stage.setKeyboardFocus(musicSelectionPage);
			}
		};
		
		ChangeListener analysisPageButtonListener = new ChangeListener() {
			@Override
			public void changed(ChangeEvent event, Actor actor) {
				setDisplayedPage(analysisPage);
			}
		};
		
		ChangeListener returnToMainPageListener = new ChangeListener() {
			@Override
			public void changed(ChangeEvent event, Actor actor) {
				setDisplayedPage(mainPage);
			}
		};
		
		ChangeListener optionsPageButtonListener = new ChangeListener() {
			@Override
			public void changed(ChangeEvent event, Actor actor) {
				setDisplayedPage(optionsPage);
			}
		};
		
		ChangeListener graphicsPageButtonListener = new ChangeListener() {
			@Override
			public void changed(ChangeEvent event, Actor actor) {
				setDisplayedPage(graphicsPage);
			}
		};
		
		ChangeListener keybindPageButtonListener = new ChangeListener() {
			@Override
			public void changed(ChangeEvent event, Actor actor) {
				setDisplayedPage(keybindPage);
			}
		};
		
		ChangeListener creditPageButtonListener = new ChangeListener() {
			@Override
			public void changed(ChangeEvent event, Actor actor) {
				setDisplayedPage(creditsPage);
			}
		};
		
		ChangeListener bloomLevelShaderListener = new ChangeListener() {
			@Override
			public void changed(ChangeEvent event, Actor actor) {
				int bloomLevel = 0;
				if (graphicsPage != null) {
					bloomLevel = graphicsPage.getBloomLevel();
				} else {
					bloomLevel = rhythmBullet.getPreferences().getInteger("glow shader");
				}
				if (bloomLevel > 0) {
					if (bloomShader == null) bloomShader = new BloomShader(screenBatch);
					bloomShader.setBloomLevel(bloomLevel);
				} else if (bloomShader != null) {
					bloomShader.dispose();
					bloomShader = null;
				}
			}
		};
		
		ChangeListener confirmedSongListener = new ChangeListener() {
			@Override
			public void changed(ChangeEvent event, Actor actor) {
				
			}
		};
	}
}
